'''
code for predictor.
'''
import importlib.resources
from tensorflow import convert_to_tensor


# package imports
from TADA_T2.backend.features import create_features, scale_features_predict
from TADA_T2.backend.model import TadaModel

def get_model_path():
    ''' 
    Quick function to get back the path to the model.
    '''
    model_weights_path = importlib.resources.files('TADA_T2.data') / 'tada.14-0.02.hdf5'
    return str(model_weights_path)

# trying to avoid recreating the model multiple times
model_cache = None  # Global variable to store the model

def predict_tada(sequences, return_both_values=False):
    '''
    Parameters
    ----------
    sequences : list
        List of sequences to predict TADA scores for.

    return_both_values : bool
        Whether to return both values. The model returns two values, one for each category
        (is a TAD or is not a TAD). If True, both values are returned. If False, only the first
        value is returned. Default is False.
        The first value matches the 'TAD' scores that are used in the TADA paper.

    Returns
    -------
    list
        List of TADA scores for each input sequence.
    '''
    global model_cache  # Use the cached model

    if not isinstance(sequences, list):
        raise Exception('Sequences must be input as a list!')

    # Defines the sequence window size and steps (stride length). Change values if needed.
    SEQUENCE_WINDOW = 5
    STEPS = 1
    LENGTH = 40

    # get scaled features
    features =create_features(sequences, SEQUENCE_WINDOW, STEPS)
    features =  scale_features_predict(features)
    features = convert_to_tensor(features)

    # check cached model
    if model_cache is None:
        # Load the model
        model_cache = TadaModel().create_model()
        # Load weights
        model_cache.load_weights(str(get_model_path()))
    # run predictions
    predictions = model_cache.predict(features, verbose=0)
    # return predictions. 
    if return_both_values:
        return predictions
    return [i[0] for i in predictions]



# various utilities
import random

def sliding_window(s: str, window_length: int, overlap: int) -> list:
    """
    Generates a list of substrings using a sliding window approach with specified overlap.

    Parameters:
    s (str): The input string.
    window_length (int): The length of the window (i.e., length of each substring).
    overlap (int): The number of characters that should overlap between consecutive substrings.

    Returns:
    list: A list of substrings of length `window_length` with specified overlap.
    """
    if window_length > len(s) or window_length <= 0:
        raise ValueError("Window length must be a positive integer and less than or equal to the length of the input string.")
    
    if overlap < 0 or overlap >= window_length:
        raise ValueError("Overlap must be a non-negative integer less than the window length.")
    
    step = window_length - overlap
    return [s[i:i + window_length] for i in range(0, len(s) - window_length + 1, step)]


def pad_sequence(input_sequence, pad='GS', 
                    objective_length=40,
                    approach='even'):
    '''
    Function to pad a sequence less than 40 amino acids. 

    Parameters
    ----------
    input_sequence : str
        The input sequence to pad.
    pad : str
        The approach to pad your sequence. 
        Options are 'random' or 'GS'.
        GS will pad both sides of your sequence with GS.
        Random will pad both sides of your sequence with random amino acids.
    objective_length : int
        The length of the sequence to pad to.
    approach : str
        The approach to pad your sequence. 
        Options are 'even' or 'N' or 'C'.
        Even will pad the sequence evenly on both sides.
        'N' will pad the sequence only at the N-terminus.
        'C' will pad the sequence only at the C-terminus.

    Returns
    -------
    str
        The padded sequence.
    '''
    if len(input_sequence) >= objective_length:
        return input_sequence
    if pad not in ['random', 'GS']:
        raise ValueError('Pad must be either random or GS.')
    if approach not in ['even', 'N', 'C']:
        raise ValueError('Approach must be either even, N, or C.')
    if objective_length < 0:
        raise ValueError('Objective length must be a positive integer.')
    num_to_pad = objective_length - len(input_sequence)
    amino_acids=['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T','V', 'W', 'Y']
    if pad=='random':
        padding=[random.choice(amino_acids) for _ in range(num_to_pad)]
    else:
        padding=[random.choice(['G', 'S']) for _ in range(num_to_pad)]
    if approach=='even':
        num_N = num_to_pad // 2
        num_C = num_to_pad - num_N
        return ''.join(padding[:num_N]) + input_sequence + ''.join(padding[num_N:])
    if approach=='N':
        return ''.join(padding)+input_sequence 
    if approach=='C':
        return input_sequence+''.join(padding)

def make_sequences_constant_length(sequence_list, objective_length=40, 
                                        overlap_length=39, pad='GS', approach='even'):
    '''
    Function to make all sequences a constant 40 amino acids using the
    pad_sequence() and sliding_window() functions as outlined above.

    Parameters
    ----------
    sequence_list : list
        The list of sequences to pad or to make windowed seqs of.
    objective_length : int
        The length of the sequence to pad to.
    overlap_length : int
        The length of the overlap between sequences.
    pad : str
        The approach to pad your sequence. 
        Options are 'random' or 'GS'.
        GS will pad both sides of your sequence with GS.
        Random will pad both sides of your sequence with random amino acids.
    approach : str
        The approach to pad your sequence. 
        Options are 'even' or 'N' or 'C'.
        Even will pad the sequence evenly on both sides.
        'N' will pad the sequence only at the N-terminus.
        'C' will pad the sequence only at the C-terminus.

    Returns
    -------
    dict:
        A dictionary with the sequence as the key and the padded or windowed sequences as the values.
    '''
    over_length = [seq for seq in sequence_list if len(seq)>objective_length]
    under_length = [seq for seq in sequence_list if len(seq)<objective_length]
    at_length = [seq for seq in sequence_list if len(seq)==objective_length]
    padded_seqs = [pad_sequence(seq, pad=pad, objective_length=objective_length, approach=approach) for seq in under_length]
    windowed_seqs = [sliding_window(seq, objective_length, overlap_length) for seq in over_length]
    over_length_dict = dict(zip(over_length, windowed_seqs))
    under_length_dict = dict(zip(under_length, padded_seqs))
    at_length_dict = dict(zip(at_length, [seq for seq in at_length]))
    return {**over_length_dict, **under_length_dict, **at_length_dict}


def map_sequences_to_prediction(sequence_dict):
    '''
    Function that takes in the sequence dict and then returns
    a list of all the values in the dict and a dict for 
    all keys in the dict with index numbers matched to the values
    in the list.

    Parameters
    ----------
    sequence_dict : dict
        A dictionary with the sequence as the key and the padded or windowed sequences as the values.

    Returns
    -------
    list, dict:
        a list of sequences and a dictionary with the sequence as the key index values for every sequence
        that match to that sequence.
    '''
    seq_list=[]
    dict_to_seq={}
    seq_index=0
    for s in sequence_dict:
        if isinstance(sequence_dict[s], list):
            seq_list.extend(sequence_dict[s])
            dict_to_seq[s]=list(range(seq_index, seq_index+len(sequence_dict[s])))
            seq_index+=len(sequence_dict[s])
        else:
            seq_list.append(sequence_dict[s])
            dict_to_seq[s]=[seq_index]
            seq_index+=1
    return seq_list, dict_to_seq


def verbose_warning_message(overlap_length, pad, approach):
    '''
    Function to make a warning message when running in verbose.
    
    Parameters
    ----------
    overlap_length : int
        The length of the overlap between sequences.
    pad : str
        The approach to pad your sequence. 
        Options are 'random' or 'GS'.
        GS will pad both sides of your sequence with GS.
        Random will pad both sides of your sequence with random amino acids.
    approach : str
        The approach to pad your sequence. 
        Options are 'even' or 'N' or 'C'.
        Even will pad the sequence evenly on both sides.
        'N' will pad the sequence only at the N-terminus.
        'C' will pad the sequence only at the C-terminus.
    '''
    if pad=='GS':
        pad_message='random selection of G and S'
    else:
        pad_message='random selection of all amino acids'
    if approach=='even':
        approach_message = 'evenly on the N and C terminus'
    elif approach=='N':
        approach_message = 'only at the N terminus'
    else:
        approach_message = 'only at the C terminus'
    warning_message=f'Warning: Not all sequences are 40 amino acids long.\nSequences shorter than 40 amino acids will be padded with a {pad_message} {approach_message}.\nSequences longer than 40 amino acids will be windowed to make sequences 40 amino acids in length with {overlap_length} overlapping amino acids.'
    return warning_message



"""Provide the primary functions."""

import protfasta
import os

from TADA_T2.backend.predictor import predict_tada as _predict_tada
from TADA_T2.backend.utils import make_sequences_constant_length, map_sequences_to_prediction, verbose_warning_message


def predict(sequences, overlap_length=39, pad='GS', approach='even', verbose=True, safe_mode=True):
    """
    Predicts TAD scores for a sequence or a list sequences.

    Parameters
    ----------
    sequences : str or list
        string of single sequence or list of sequences to predict TADA scores for.

    overlap_length : int
        The length of the overlap between sequences.
        Default is 39
    
    pad : str
        The approach to pad your sequence. 
        Options are 'random' or 'GS'.
        GS will pad both sides of your sequence with GS.
        Random will pad both sides of your sequence with random amino acids.
        Default is 'GS'.
    
    approach : str
        The approach to pad your sequence. 
        Options are 'even' or 'N' or 'C'.
        Even will pad the sequence evenly on both sides.
        'N' will pad the sequence only at the N-terminus.
        'C' will pad the sequence only at the C-terminus.
        Default is 'even'.

    verbose : bool
        whether to warn user when sequence lengths are not 40 amino acids.

    safe_mode : bool
        whether to run the function in safe mode. Safe mode will raise an exception
        if any sequences are under 40 amino acids. Default is True.

    Returns
    -------
    dict
        A dict with the sequence as the key and the scores as the values.
    """
    if isinstance(sequences, str):
        sequences=[sequences]

    # get sequence lengths
    seq_lengths=[len(seq) for seq in sequences]
    if safe_mode:
        # check if all sequences are over 40 amino acids long   
        if not all([length>=40 for length in seq_lengths]):
            raise ValueError('Not all sequences are 40 amino acids long. TADA was not made for sequences under 40 amino acids. You can still make these predictions by setting safe_mode=False, but use this feature with extreme caution!.')        
    if any([length!=40 for length in seq_lengths]):
        if verbose:
            message = verbose_warning_message(overlap_length=overlap_length, pad=pad, approach=approach)
            print(str(message))
    seq_dict=make_sequences_constant_length(sequences, 
                                            overlap_length=overlap_length, 
                                            pad=pad, approach=approach)
    padded_or_trimmed_seqs, map_to_predictions=map_sequences_to_prediction(seq_dict)
    predictions=_predict_tada(padded_or_trimmed_seqs)
    # holds final sequences
    final_dict={}
    # map the indices in the predictions to the original sequences
    for seq, indices in map_to_predictions.items():
        final_dict[seq]=[[padded_or_trimmed_seqs[index],predictions[index]] for index in indices]
    return final_dict


def predict_from_fasta(path_to_fasta, overlap_length=39, pad='GS', 
                        approach='even', verbose=True, safe_mode=True):
    """
    Predicts TAD scores for sequences in a .fasta file

    Parameters
    ----------
    path_to_fasta : str
        path to a .fasta file as a string

    overlap_length : int
        The length of the overlap between sequences.
        Default is 39
    
    pad : str
        The approach to pad your sequence. 
        Options are 'random' or 'GS'.
        GS will pad both sides of your sequence with GS.
        Random will pad both sides of your sequence with random amino acids.
        Default is 'GS'.
    
    approach : str
        The approach to pad your sequence. 
        Options are 'even' or 'N' or 'C'.
        Even will pad the sequence evenly on both sides.
        'N' will pad the sequence only at the N-terminus.
        'C' will pad the sequence only at the C-terminus.
        Default is 'even'.

    safe_mode : bool
        Whether to run the function in safe mode. Safe mode will raise an exception
        if any sequences are under 40 amino acids. Default is True.

    verbose : bool
        whether to warn user when sequence lengths are not 40 amino acids.

    Returns
    -------
    dict
        Dict with name of the sequence from the fasta file as the key
        and then a list as the values where the first element is the sequence
        and the second element is the TADA score. 
    """
    # make sure that path is valid
    if not os.path.exists(path_to_fasta):
        raise ValueError('Path does not exist.')        

    # read in sequences
    sequences=protfasta.read_fasta(path_to_fasta, invalid_sequence_action='convert')
    seq_names=list(sequences.keys())
    sequences=list(sequences.values())

    if safe_mode:
        # check if all sequences are over 40 amino acids long
        seq_lengths=[len(seq) for seq in sequences]
        if not all([length>=40 for length in seq_lengths]):
            raise ValueError('Not all sequences are 40 amino acids long. TADA was not made for sequences under 40 amino acids. You can still make these predictions by setting safe_mode=False, but use this feature with extreme caution!.')
    
    # run predictions
    predictions=predict(sequences, overlap_length=overlap_length, 
                        pad=pad, approach=approach, verbose=verbose,
                        safe_mode=safe_mode)

    # map sequence names to predictions
    final_dict={}
    for i, s in enumerate(seq_names):
        final_dict[s]=[sequences[i], predictions[sequences[i]]]
    return final_dict



'''
Code for the encoding features. Modified to be compaitble with Tensorflow2 along
with optimizations to make it faster.
'''
from copy import deepcopy
import importlib.resources

import numpy as np
import alphaPredict as alpha
from localcider.sequenceParameters import SequenceParameters


def get_scaler_path():
    ''' 
    Quick function to get back the path to the model.
    '''
    scaler_arr_path = importlib.resources.files('TADA_T2.data') / 'scaler_metric.npy'
    return str(scaler_arr_path)


def create_features(sequences, SEQUENCE_WINDOW = 5, STEPS = 1, LENGTH = 40, PROPERTIES = 42):
    '''
    Function to create features for the model. Updated to improve readability.

    Parameters
    ----------
    sequences : List. 
        List of sequences with max length of 40AA.
    SEQUENCE_WINDOW : Int, optional
        The window that the sequence is scanned over.
    STEPS : Int, optional
        The size of steps. 
    LENGTH : int, optional
        The length of the sequences. 
    PROPERTIES : Int, optional
        The number of properties
    
    Returns
    -------
    features : Processed input data for model
    '''

    aliphatics_set = set(['I', 'V', 'L', 'A'])
    aromatics_set = set(['W', 'F', 'Y'])
    branching_set = set(['V', 'I', 'T'])
    charged_set = set(['K', 'R', 'H', 'D', 'E'])
    negatives_set = set(['D', 'E'])
    phosphorylatables_set = set(['S', 'T', 'Y'])
    polars_set = set(['R', 'K', 'D', 'E', 'Q', 'N', 'Y'])
    hydrophobics_set = set(['W', 'F', 'L', 'V', 'I', 'C', 'M'])
    positives_set = set(['K', 'R', 'H'])
    sulfurcontaining_set = set(['M', 'C'])
    tinys_set = set(['G', 'A', 'S', 'P'])
    amino_acids = ['R', 'K', 'D', 'E', 'Q', 'N', 'H', 'S', 'T', 'Y', 'C', 'W', 'M', 'A', 'I', 'L', 'F', 'V', 'P', 'G']

    features = []        
    for sequence in sequences:
        SEQUENCE_LENGTH = len(sequence)
        SeqOb = SequenceParameters(sequence)
        kappa = np.full(int((SEQUENCE_LENGTH-SEQUENCE_WINDOW)/STEPS+1), SeqOb.get_kappa())
        omega = np.full(int((SEQUENCE_LENGTH-SEQUENCE_WINDOW)/STEPS+1), SeqOb.get_Omega())
        
        # make subseqs.
        sub_seq = [sequence[STEPS * j:STEPS * j + SEQUENCE_WINDOW] for j in range((SEQUENCE_LENGTH - SEQUENCE_WINDOW) // STEPS + 1)]
        one = np.array([sum(aa in aliphatics_set for aa in seq) for seq in sub_seq])
        two = np.array([sum(aa in aromatics_set for aa in seq) for seq in sub_seq])
        three = np.array([sum(aa in branching_set for aa in seq) for seq in sub_seq])
        four = np.array([sum(aa in charged_set for aa in seq) for seq in sub_seq])
        five = np.array([sum(aa in negatives_set for aa in seq) for seq in sub_seq])
        six = np.array([sum(aa in phosphorylatables_set for aa in seq) for seq in sub_seq])
        seven = np.array([sum(aa in polars_set for aa in seq) for seq in sub_seq])
        eight = np.array([sum(aa in hydrophobics_set for aa in seq) for seq in sub_seq])
        nine = np.array([sum(aa in positives_set for aa in seq) for seq in sub_seq])
        ten = np.array([sum(aa in sulfurcontaining_set for aa in seq) for seq in sub_seq])
        eleven = np.array([sum(aa in tinys_set for aa in seq) for seq in sub_seq])
        sstructure = np.array([sum(alpha.predict(seq)) / len(seq) for seq in sub_seq])
        count_20 = np.array([[s.count(aa) for s in sub_seq] for aa in amino_acids])
        
        # turn subseqs into SeqObs.
        sub_seq = [SequenceParameters(seq) for seq in sub_seq]
        hydropathy = np.array([seq.get_mean_hydropathy() for seq in sub_seq])
        hydropathy_ww = np.array([seq.get_WW_hydropathy() for seq in sub_seq])
        ncpr = np.array([seq.get_NCPR() for seq in sub_seq])
        promoting = np.array([seq.get_fraction_disorder_promoting() for seq in sub_seq])
        fcr = np.array([seq.get_FCR() for seq in sub_seq])
        charge = np.array([seq.get_mean_net_charge() for seq in sub_seq])
        negative = np.array([seq.get_fraction_negative() for seq in sub_seq])
        positive = np.array([seq.get_fraction_positive() for seq in sub_seq])

        # make array of features.
        x = np.array([kappa, omega, hydropathy, hydropathy_ww, ncpr, promoting, fcr, charge, negative, positive,
                     one, two, three, four, five, six, seven, eight, nine, ten, eleven, sstructure])
        # concatenate x and count_20
        x = np.concatenate([x, count_20])
        # do padding 
        padded_x = np.zeros((PROPERTIES, (LENGTH - SEQUENCE_WINDOW) // STEPS + 1))
        padded_x[:x.shape[0], :x.shape[1]] = x
        # append to features
        features.append(padded_x)

    features = np.array(features)
    return np.transpose(features, (0, 2, 1))


def scale_features_predict(features: np.ndarray, SEQUENCE_WINDOW=5, STEPS=1, LENGTH=40) -> np.ndarray:
    '''
    Function to scale the features for prediction. Updated to improve readability.

    Parameters
    ----------
    features : np.ndarray
        Takes the output of create_features() and scales the values per feature column.

    Returns
    -------
    scaled_array_copy : np.ndarray
        Scaled feature array ready for prediction.
    '''

    scaled_array_copy = deepcopy(features)
    n, m = features[0].shape
    scaler_metric = np.load(str(get_scaler_path()))


    # Precompute the length for reshaping
    num_steps = int((LENGTH - SEQUENCE_WINDOW) / STEPS + 1)
    
    for i in range(m):
        # Load StandardScaler and MinMaxScaler parameters for each feature
        mean_, var_, scale_, n_samples_seen_ = scaler_metric[i, :4]
        min_, data_min_, data_max_, scale2_, n_samples_seen2_, data_range_ = scaler_metric[i, 4:10]
        
        # Normalize using StandardScaler metrics
        results = (scaled_array_copy[:, :, i] - mean_) / scale_

        # Apply MinMaxScaler metrics
        results = (results - data_min_) / data_range_

        # Reshape back to original dimensions
        results = results.reshape((len(features), num_steps))
        scaled_array_copy[:, :, i] = results

    return scaled_array_copy





'''
Code for the ML Model. Modified to be compaitble with Tensorflow2 along
with a few other optimizations. 
Credit to Lisa (see https://github.com/LisaVdB/TADA)
for original implementation of the model.
'''
import importlib.resources

from tensorflow import nn, matmul, reduce_sum, constant
from tensorflow.keras import layers, regularizers, metrics, models, optimizers
from tensorflow.keras.models import Sequential

class Attention(layers.Layer):
    '''
    Custom Attention class. Updated to be a bit more 
    efficient and compatible with Tensorflow2. 
    '''
    def __init__(self, return_sequences=True, **kwargs):
        super(Attention, self).__init__(**kwargs)
        self.return_sequences = return_sequences

    def build(self, input_shape):
        # Weight matrix for attention
        self.W = self.add_weight(
            name="att_weight", shape=(input_shape[-1], 1), 
            initializer="glorot_uniform", trainable=True
        )
        # Bias term
        self.b = self.add_weight(
            name="att_bias", shape=(input_shape[1], 1), 
            initializer="zeros", trainable=True
        )
        super(Attention, self).build(input_shape)

    def call(self, inputs):
        # Compute the attention scores
        e = nn.tanh(matmul(inputs, self.W) + self.b)
        # Softmax over the attention scores along the time axis
        a = nn.softmax(e, axis=1)
        # Apply attention weights to the input
        output = inputs * a
        
        if self.return_sequences:
            return output
        # If not returning sequences, sum the weighted input along the time axis
        return reduce_sum(output, axis=1)

    def get_config(self):
        config = super().get_config()
        config.update({"return_sequences": self.return_sequences})
        return config


class TadaModel:
    def __init__(self, 
                 shape=(36,42), 
                 kernel_size=2, 
                 filters=100, 
                 activation_function='gelu', 
                 learning_rate=1e-3, 
                 dropout=0.3, 
                 bilstm_output_size=100):
        """
        Initialize the CustomModel and build the architecture.
        """
        self.shape = shape
        self.kernel_size = kernel_size
        self.filters = filters
        self.activation_function = activation_function
        self.learning_rate = learning_rate
        self.dropout = dropout
        self.bilstm_output_size = bilstm_output_size
        self.model = self.create_model()

    def create_model(self):
        """
        Define the NN architecture.
        """
        model = Sequential()
        
        # Add an explicit Input layer
        model.add(layers.Input(shape=self.shape))

        model.add(layers.Conv1D(filters=self.filters, 
                                 kernel_size=self.kernel_size,
                                 padding='valid',
                                 activation=self.activation_function,
                                 strides=1,
                                 kernel_regularizer=regularizers.l1_l2(l1=1e-5, l2=1e-4)))
        
        model.add(layers.Dropout(self.dropout)) 
        model.add(layers.Conv1D(filters=self.filters,
                                 kernel_size=self.kernel_size,
                                 padding='valid',
                                 activation=self.activation_function,
                                 strides=1))
        model.add(layers.Dropout(self.dropout))
        
        model.add(Attention())  # Custom attention layer
        
        model.add(layers.Bidirectional(layers.LSTM(self.bilstm_output_size, return_sequences=True)))  # Bidirectional LSTM
        model.add(layers.Bidirectional(layers.LSTM(self.bilstm_output_size))) 
        model.add(layers.Dense(2, activation="softmax"))  # Output layer
        
        return model




"""
Unit and regression test for the TADA_T2 package.
"""

# Import package, test suite, and other packages as needed
import sys

import pytest

import TADA_T2


def test_TADA_T2_imported():
    """Sample test, will always pass so long as import statement worked."""
    assert "TADA_T2" in sys.modules



'''
Makes sure the predictor gets the same values as we expect from the original predictor. 
'''
from TADA_T2.TADA import predict


def is_close(a, b, tolerance=0.000001):
    '''
    Function to check if two numbers are close. 
    '''
    return abs(a-b) <= tolerance


def test_predictor_is_correct():
    '''
    Function to make sure this version of TADA gets the same TAD predictions as 
    the original version of TADA. This tests 100 sequences. 
    '''
    original_TADA_scores={'QFNENSNIMQQQPLQGSFNPLLEYDFANHGGQWLSDYIDL': 0.637078643, 'EFSPENSSSSSWSSQESFLWEESFLHQSFDQSFLLSSPTD': 0.645964801, 'VLPPLSESFDLDSLMSTPMSSPRQNSIEAETNSSTFFDFG': 0.662755132, 'SWLLPNSGKNSGNNNGFSIGDEFLNLVDYSSSDKQFTDQS': 0.577608645, 'QAFGNSPQNSSSNGSLSSSLDEENNFFFSLTSEEHNKSNN': 0.4973737, 'ELFKNHNEDYIQTQYGTNDADEYMSKFLDSFLDIPYEPEQ': 0.602484941, 'IVQQEDLSEWEGFNADTFFSDNNNNYNLNVHHQLTPYGDG': 0.576722324, 'NATSDLNMDQDYDFSQFFEKFGGDNHNEENSMNDLLMSDV': 0.676980257, 'KNSINVFGGEHGYEDFGFCYDDKFSSFLNSLINDVGDPFG': 0.646433651, 'IMEDGVIDEIHKQSDLPLWYDDLITTDEDPLMSSILGDLL': 0.670372307, 'ELRFEYIEEAYSEFNDIIIQEVDKPDLLEIPFDSDPDIWS': 0.691135406, 'PTEEQNPFFLPDLFRSGDYFWDSEITPDPLFLDEFHQSLL': 0.671409309, 'LKKNNVCENSITCNKDDEKDDFVNNLMNGDNMWLENLLGE': 0.575095773, 'ANSSSSSQQVFEFEYLDDSVLDELLEYGENYNKTHNINMG': 0.630013287, 'DYGWPNDVDQSHLDSSDMFDVDELLRDLNGDDVFAGLNQD': 0.641659141, 'CFIPEMDMIIPETDSFFFQSQPQLEFHQPLFQEEAPSQTH': 0.641651869, 'SDLDFLLDDENGDFADFDFSFDNSDDFFDFDLSEPAVVIP': 0.688971817, 'DGLMNPNDAFCLCNGTFTWQLYGEEDVGFRFEEPFNWQND': 0.662678242, 'DQIFCENDDIFNDMLFLGGETMNIEDELTSSSIKDMGSTF': 0.650848985, 'NQLNNIMDPSTTLNQITSDIWFEDDQAILFDQQQSFSGAF': 0.630743623, 'KEEDMFSGSDDYLIQMEHDDGWLHEAMSNLIPFPCEFDAP': 0.694395065, 'ILPIQEDLPLVWSLEDLDSLLSGTELHKLVKEDHVLIYED': 0.533637226, 'EQVEEETKKLKASGCFDRSLHDFDEIQHMDDMFLSSILED': 0.612104237, 'ESSSTVTDQNSSMDNENHLIDNIYDDDELFSYLWSDETTK': 0.625922024, 'MCALVPPLYPNFGWPCGDHSFYETDDVSNTFLDFPLPDLT': 0.660090864, 'CKVDYLSYQGDDLSSILYRIDESDFTFEGLRMDGHDQLGE': 0.577952623, 'FSDDLLDGIDYYDDLFIGFDGDDVLPDLEIDSEILGEYSG': 0.718448281, 'RDENLNGGDLLSFEFDGSHFFNFSIFDHETTCKRLKRSSE': 0.552880168, 'MHTKPELHEVNGLNEIQFLLDHDDFDDITSEFLQDNDILF': 0.669813037, 'EKPGQENFFGMSVDDLGTPKNEQEDFSLWDVLDPDMLFSD': 0.678077281, 'QQQQDSLSVADYGWPNDVDQSHLDSSDMFDVDELLRDLNG': 0.633327305, 'FEFQGLLDDKEIQEILECSFSEEPDQLVSQGSFMINGDNW': 0.610447705, 'MNSIFSIDDFSDPFWETPPIPLNPDSSKPVTADEVSQSQP': 0.528875291, 'WDYEFDYDDEIENGTEVEPVEIGDGVDWWKIDTEDGVGSS': 0.600601971, 'YNYSLNSIPDAENDLSFFDNGDKEKNDLFYGWGDIGNFED': 0.669441521, 'PPRQPRHVGDGVAFGQFLDLGSSGQIDFDAAAAAFFPNLP': 0.587494075, 'ENQEADAIVPEATTAEHGATLAFDVEQLWSLFDGETVELD': 0.605891705, 'AENDLSFFDNGDKEKNDLFYGWGDIGNFEDVDNMLRSCDS': 0.660498917, 'ASILWAEEEDMCNEKSHVLTYADKEGDWMMVGDVPWEMFL': 0.63095367, 'CHSPPPWNEQEETGSPFRTENFSWDTLIEMPRSETTTMQF': 0.483799428, 'QFQMEEFDPFYQSSEHIIDHMKEDISINNSEYDFSQFLEQ': 0.674408555, 'FANVQGESQIDDATTPIEEEWKTWLNNDGDEQRNIMFMQD': 0.550652266, 'DQTPIGELMSDLGFPDGEFELTFDGMDDLYFPAENESFLI': 0.702092111, 'AEDTRDYHNQDGNWLDYLWFERLHDLNFSDQGFEDQTSTV': 0.647073627, 'GVDYWFRSEVGEVSITDMWPDESGPDWNQMITFDQDHAGP': 0.613302588, 'EGNGGFSPNSSFGAFADTAMDLDFMDELLFDGCWLETTDS': 0.671220005, 'SMDELDFNKDFDLPPSSNQTLGLANGFYLDDLDFSSLDPP': 0.697582006, 'TSLNHPSTAQHSSGSDFLEDWEKFLDDETSDSCWKSFLDL': 0.640011549, 'AGATFDYPFLEALQEIIDSSSSSPPLILQNGQEENFNNPM': 0.584760666, 'TSVNTPRDRLCTVDDLQLHIGDWFYTDGAGQEQGPLSFSE': 0.604496717, 'MEMESFMDDLLNFSVPEEEEDDDEHTQPPRNITRRKTGLR': 0.361037374, 'EVSKQGENETEDFEFGLIDDFESSPWDVDHFFDHHHHSFD': 0.654456675, 'GSSDSEGKFGFQTGPDFTLEEFQKYDEYFKECYFQSEDHP': 0.610997081, 'QEPLEFGVDETFDINELLGILNDNNVSGQETMQYQVDRHP': 0.577261984, 'DAEIDAGFQFLAPDLFSTCELESGLKWFDQQDHEDFPYCS': 0.697733223, 'MEGRVNALSNINDLELHNFLVDPNFDQFINLIRGDHQTID': 0.568151653, 'TMVDEIPMSVPSLMTGLSQDDDFVPWLNHHPSLDGYCSDF': 0.61768049, 'PEMSIFDQPMDQIFCENDDIFNDMLFLGGETMNIEDELTS': 0.702756763, 'FSQSYEYMIDNKEDLGTSIDLNIPEYDFPQFLEQLINDDD': 0.682582974, 'SCESNNVKNSEPYGGMSVGHKNIETMADDFVDWDFVWREG': 0.56252414, 'SDGAREERPLYLFDPKFAEKVPVLDSEYDVPVYFREDLFG': 0.546336293, 'SWSSQESFLWEESFLHQSFDQSFLLSSPTDNYCDDFFAFE': 0.718793273, 'DQHYYHLAFFTTRDIEAMEELAWDYGIDFNDNDSLMKPFD': 0.611177921, 'MIGTSFPEDLDCGNFFDNMDDLMDFPGGDIDVGFGIGDSD': 0.687168658, 'RTMVSSKEVLSKCEWPDWADQLISDDSLEPNWSELLGDPN': 0.631952882, 'FNLDSILASENGSLMDGSFNAESYHQLQQWPFDGYHQPEW': 0.584000826, 'EVTKSQSFDHPQPDIPCGFEDTNEESDLRRQLVESTTPNN': 0.205637708, 'VGGASLSGGGDTPKTTSSQIFNEDTLDQFLELMGRSCKEE': 0.479624182, 'QEQQQQQLQPDLLTVADYGWPWSNDIVNDQTSWDPNECFD': 0.660169959, 'DYNDVSEFGFAAETTSDGLPDVCWEQFAAGITETGFNWPT': 0.64871031, 'QRDEDPLYVFDDKFGEAAPELLKDYSVPHLFQEDWFEILD': 0.689492226, 'EDYFRRLDFPDLETLSDLASLRSLSSRNCFSIPSVEYDSI': 0.452268034, 'NLSWFENINGAASSSDSLWNIGETDEEFWFLQQQQQFNNN': 0.645190895, 'AESSLVISGDSHSDACDEATTAELVDFKWYPELESLDFTL': 0.623853266, 'ETNDLSCAQQDDFNFEDYLSFFDDEGLTFDDSLLMGPEDF': 0.708377421, 'DLGFPDGEFELTFDGMDDLYFPAENESFLIPINTSNQEQF': 0.680244625, 'APLVTKAKGRKLTAEELWSELDASAADDFWGFYSTSKLHP': 0.542910457, 'SPIFSSEKKTAVSGADDVAVFFPMGEEDESLFADLGELPE': 0.592305183, 'KESNEETVFDLPDLFTDGLMNPNDAFCLCNGTFTWQLYGE': 0.63344568, 'CYNDFPVESNYLIGEAFLDPNSNLLENDGLYLETNDLSST': 0.686418056, 'ETEESLNCCVPVFDPFSDMLIDDINGFCLVPDEVNNTTTN': 0.656974971, 'DAVQELWKKKDVVDKELIASLRAAAGLPTEEEIFSIFPFS': 0.446763277, 'LPDFDFADVEDLQLADSSFGFLDQLAPINISCPLKSFAAS': 0.598931491, 'MERNNRNEGTHEEEQCSLSDIIYSFCSENHSELNPLQEIF': 0.550824165, 'GDDIASSSSRRKTPFQFDLNFPPLDGVDLFAGGIDDLHCT': 0.591349363, 'MDHSSVPENSSMAKELGVCEDDFNGNLISDEVDLALENYE': 0.570704162, 'VSDENTLHRSNDFSTADFHTSGLSVPMDDIAELEWLSNFV': 0.631526768, 'EPKWGELEDALEAFDTSMFGSSMELLQPDAFVPQFLYQSD': 0.65042305, 'LLVCDGFMGDFDFDASFVSGLWCIEPHVPKQEPDSPVLDP': 0.594468057, 'VTVVEQEANHNGDVFLDRFNEALHYYSSLFDSLEDGVVIP': 0.596358716, 'IENVLMNTFDGAEAAVLDLKQYNPEEDLILTELIDELARD': 0.596449316, 'EEEEEDHLLSSTFESINGHSRDQHNHSIDDFESIFDITID': 0.621232986, 'ENENIVGPEQDLLMSDFPSTFVDEDDILGDITSWSTYLLD': 0.678622127, 'NFGWPCGDHSFYETDDVSNTFLDFPLPDLTVTHENVSSEN': 0.635302722, 'VRDVLEGASDLTFEFPDFEECMNLIRGRKSLSDDIVLKSI': 0.470604807, 'SMIDEGHLWGSLWNLDEDDPHSFGGGSGQGTAADIDEKFP': 0.551305234, 'SFIDQNHLYPLPNISTVEDISFLEYNVDKTENSGSEKLAN': 0.497447371, 'FMQSSSSPEIHQFEDLFKSYKLSDEMNNLVEASEYDFGEE': 0.579854369, 'SKAFYGASSLHDFEGIEQMDDMFLSSILEDVPEDDGDVHR': 0.615141034, 'YRFHPTDEELVDYYLKNKVAFPGMQVDVIKDVDLYKIEPW': 0.256095409}
    sequences=list(original_TADA_scores.keys())
    TADA_scores=predict(sequences)
    # assert each score is equivalent. 
    for seq in sequences:
        assert is_close(original_TADA_scores[seq], TADA_scores[seq][0][1])
